﻿//*****************************************************************************
//* Copyright (c) 2008-2011 Miday Workshop All rights reserved.
//*
//* File: ExString.cpp
//*****************************************************************************

#ifdef _WINDOWS_PLAT
#include <windows.h>
#else  //_WINDOWS_PLAT
#include <stdint.h>
#endif  //_WINDOWS_PLAT
#include <new>
#include <stdio.h>
#include <string.h>
#include <wchar.h>
#include <memory.h>
#include "CalcDef.h"
#include "MsgLocale.h"

//==============================================================================
//構造体を動的確保する（領域は、指定値に+1バイトした値とする）
//==============================================================================
int CharLenNewAlloc(CHAR_LEN *char_len, int size) {
	int irc;
	try{
		char_len->size = size;
		if ( char_len->size <= 0 ) char_len->size = 100;
		char_len->data = new UNICHAR [char_len->size+1];
		char_len->data[0] = LIT_NULL_C;
	}
#if _MSC_VER > 1200
	catch (std::bad_alloc) {
#else  //MSC
//VC6 以下
	catch (...) {
#endif  //MSC
		irc = MessageDisp( MSG_ER_MEM2_CODE );
		return MEMORY_ERROR;
	}
#if _MSC_VER > 1200
	catch (...) {
		irc = MessageDisp( MSG_ER_MEM3_CODE );
		return MEMORY_ERROR;
	}
#endif  //MSC
	return 0;
}

//==============================================================================
//動的確保された構造体を拡張する（領域は、+1バイトした値とする）
//==============================================================================
int CharLenAddAlloc(CHAR_LEN *char_len, int addsize) {
	int size;
	UNICHAR *wk;

	size = char_len->size + addsize;

	if ( size > char_len->size ) {
		try{
			wk = new UNICHAR [size+1];
		}
#if _MSC_VER > 1200
		catch (std::bad_alloc) {
#else  //MSC
//VC6 以下
		catch (...) {
#endif  //MSC
			return MEMORY_ERROR;
		}
#if _MSC_VER > 1200
		catch (...) {
			return MEMORY_ERROR;
		}
#endif  //MSC
		MTRCPY(wk, size+1, char_len->data);
		char_len->size = size;
		delete[] char_len->data;
		char_len->data = wk;
	}
	return 0;
}

//==============================================================================
//動的確保された構造体に文字列を追加する
//==============================================================================
int CharLenCat(CHAR_LEN *char_len, UNICHAR *add) {
	int size, irc;
	size = MTRLEN(char_len->data) + MTRLEN(add);

	if ( size > char_len->size ) {
		if ( CharLenAddAlloc(char_len, MTRLEN(add)+500) != 0 ) {
			irc = MessageDisp( MSG_ER_MEM3_CODE );
			return MEMORY_ERROR;
		}
	}
	MTRCAT(char_len->data, char_len->size+1, add);
	return 0;
}

//==============================================================================
//動的確保された構造体に文字列を指定バイトだけ追加する　最後にNULLを付ける
//==============================================================================
int CharLenNCat(CHAR_LEN *char_len, UNICHAR *add, int len) {
	int size, newsize, irc;
	size = len;
	if ( size > (int)MTRLEN(add) ) {
		size = MTRLEN(add);
	}
	newsize = MTRLEN(char_len->data) + size;

	if ( newsize > char_len->size ) {
		if ( CharLenAddAlloc(char_len, size+500) != 0 ) {
			irc = MessageDisp( MSG_ER_MEM3_CODE );
			return MEMORY_ERROR;
		}
	}
	MTRNCAT(char_len->data, char_len->size+1, add, size);
	char_len->data[newsize] = LIT_NULL_C;
	return 0;
}

//==============================================================================
//動的確保された構造体を削除する
//==============================================================================
int CharLenFree(CHAR_LEN *char_len) {
	if ( char_len->size != 0 ) {
		delete[] char_len->data;
	}
	char_len->size = 0;
	return 0;
}

//==============================================================================
//sprintfのラッパー
//==============================================================================
#ifdef _WINDOWS_PLAT
_CRTIMP int __cdecl SPRINTF(UNICHAR * a, int len, const UNICHAR *c, ...) {
//変数は最大5個まで指定できる
//指定できるフォーマットは、%d %x %c %s のみ
	va_list list;
	int ir, ia, ib;
	int para[100];

	a[0] = '\0';
	ia = 0;
	va_start(list, c);
	for(ib=0; ib<(int)MTRLEN(c); ib++) {
		if ( c[ib] == '%' ) {
			if ( c[ib+1] == 's' ) {
				para[ia] = (int)va_arg(list, char*);
				ia++;
			} else if ( c[ib+1] == 'c' ) {
				para[ia] = va_arg(list, char);
				ia++;
			} else if ( c[ib+1] == 'x' ) {
				para[ia] = va_arg(list, BYTE);
				ia++;
			} else {
				para[ia] = va_arg(list, int);
				ia++;
			}
		}
		if ( ia > 5 ) {
			MessageDisp( MSG_ER_SYSTEM_CODE );
			va_end(list);
			exit(0);
		}
	}
	if ( ia != 0 ) {
		switch( ia ) {
		case 1:
#if !defined(_MSC_VER) || _MSC_VER < 1400
			ir= sprintf(a, c, para[0]);
#else  //MSC
#ifdef _UNICODE_TABLE
			ir= swprintf_s(a, len, c, para[0]);
#else  //_UNICODE_TABLE
			ir= sprintf_s(a, len, c, para[0]);
#endif  //_UNICODE_TABLE
#endif  //MSC
			break;
		case 2:
#if !defined(_MSC_VER) || _MSC_VER < 1400
			ir= sprintf(a, c, para[0], para[1]);
#else  //MSC
#ifdef _UNICODE_TABLE
			ir= swprintf_s(a, len, c, para[0], para[1]);
#else  //_UNICODE_TABLE
			ir= sprintf_s(a, len, c, para[0], para[1]);
#endif  //_UNICODE_TABLE
#endif  //MSC
			break;
		case 3:
#if !defined(_MSC_VER) || _MSC_VER < 1400
			ir= sprintf(a, c, para[0], para[1], para[2]);
#else  //MSC
#ifdef _UNICODE_TABLE
			ir= swprintf_s(a, len, c, para[0], para[1], para[2]);
#else  //_UNICODE_TABLE
			ir= sprintf_s(a, len, c, para[0], para[1], para[2]);
#endif  //_UNICODE_TABLE
#endif  //MSC
			break;
		case 4:
#if !defined(_MSC_VER) || _MSC_VER < 1400
			ir= sprintf(a, c, para[0], para[1], para[2], para[3]);
#else  //MSC
#ifdef _UNICODE_TABLE
			ir= swprintf_s(a, len, c, para[0], para[1], para[2], para[3]);
#else  //_UNICODE_TABLE
			ir= sprintf_s(a, len, c, para[0], para[1], para[2], para[3]);
#endif  //_UNICODE_TABLE
#endif  //MSC
			break;
		case 5:
#if !defined(_MSC_VER) || _MSC_VER < 1400
			ir= sprintf(a, c, para[0], para[1], para[2], para[3], para[4]);
#else  //MSC
#ifdef _UNICODE_TABLE
			ir= swprintf_s(a, len, c, para[0], para[1], para[2], para[3], para[4]);
#else  //_UNICODE_TABLE
			ir= sprintf_s(a, len, c, para[0], para[1], para[2], para[3], para[4]);
#endif  //_UNICODE_TABLE
#endif  //MSC
			break;
		}
	}
	va_end(list);
	return ir;
}
#endif  //_WINDOWS_PLAT
