﻿//*****************************************************************************
//* Copyright (c) 2011-2013 Miday Workshop All rights reserved.
//*
//* File: GtkMain.cpp
//*****************************************************************************

#ifdef _LINUX_PLAT
//#include <stdio.h>
#include <sal/main.h>

#include <gtk/gtk.h>
#include <unistd.h>
#include <time.h>
#include <wchar.h>
#include <string.h>
#include <locale.h>

#include "CalcDef.h"
#include "MsgLocale.h"

const gchar *clip_key1 = "application/x-openoffice-objectdescriptor-xml";
const gchar *clip_key2 = "calc";

typedef struct _CallbackData {
	GtkWidget *window;
	GtkWidget *pbar;
} CallbackData;

CallbackData *gdata;
GtkWidget *window;

int close_button;

UNICHAR *clipdata;//クリップボードへ出力するデータのポインタ

enum {
	TARGET_STRING,
	TARGET_TEXT,
	TARGET_COMPOUND_TEXT
};


static GdkAtom atom_targets, atom_compound, atom_text;

extern struct tm *Just1, *Just2;

//==============================================================================
//setlocaleの引数を wchar_t* にしたラッパー
//==============================================================================
char* wsetlocale(int category, wchar_t* locale) {
	char *c, *p;
	int sizew, sizew2;
	sizew = wcstombs( NULL, locale, 0);
	if ( sizew == -1 ) return NULL;
	try{
		c = new char[sizew+1];
	}
	catch (...) {
		return NULL;
	}
	sizew2 = wcstombs( c, locale, sizew);
	c[sizew] = '\0';

	p = setlocale(category, c);

	delete[] c;

	return p;
}

//==============================================================================
//ProgressBarの進捗を設定する
//==============================================================================
int ProgressView(int par) {
	GtkAdjustment *adj;

	adj = GTK_PROGRESS (gdata->pbar)->adjustment;
	if (par > adj->upper) {
		par = adj->lower;
	}

	/* 新しい値を設定する */
	gtk_progress_set_value (GTK_PROGRESS (gdata->pbar), par);

	Window_Refresh();

	return 0;
}

//==============================================================================
//割り当てられたメモリを開放する
//==============================================================================
void destroy_progress( GtkWidget *widget, CallbackData *gdata) {

	if ( Out_alloc == 1 ) CharLenFree(&out_data);
	Out_alloc        = 0;

	window = NULL;
	g_free(gdata);
	gtk_main_quit();
}

//==============================================================================
//画面のリフレッシュ
//==============================================================================
int Window_Refresh() {
	while (gtk_events_pending ()) {
		gtk_main_iteration ();
	}
	return 0;
}

//==============================================================================
//終了のコールバック
//==============================================================================
int cancel_message( GtkWidget *widget, CallbackData *gdata ) {
//	gint result;

//printf("END\n");
	if ( Break_flag == 2 ) return FALSE;
	if ( close_button == 0 ) {//×ボタンが押された場合は、0
		Break_flag = 1;
	}
	return TRUE;
}

//==============================================================================
//キャンセルボタンのコールバック
//==============================================================================
int button_push_message( GtkWidget *widget, CallbackData *gdata ) {
	GdkEvent ev;

//printf("BUTTON\n");
	if ( Break_flag != 2 ) Break_flag = 1;
	//GDK_DELETEのイベントを送る
	ev.any.type = GDK_DELETE;
	ev.any.window = (widget)->window;
	ev.any.send_event = FALSE;
	gdk_event_put (&ev);
	return FALSE;
}

//==============================================================================
//ウインドウを閉じて終了する
//==============================================================================
int WinClose() {
		close_button = 1;//×ボタンは押されていない
		Break_flag   = 2;//キャンセル待ちではない
		button_push_message( window, gdata );
}

//==============================================================================
// ペーストされて、セレクションのデータが必要になったとき呼ばれる
//==============================================================================
static void signal_selection_out(GtkWidget        *widget,
                                 GtkSelectionData *data,
                                 guint             info,
                                 guint             time,
                                 gpointer          user_data) {
	char *clipwork;

	int sizew2 = wcstombs( NULL, clipdata, 0);//-1が返ってきたら変換不可
	if ( sizew2 <= 0 ) {
		MessageDisp(MSG_ER_CNV_CODE);
//		button_push_message( window, gdata );
		exit(0);
	} 
	try{
		clipwork = new char[sizew2+1];
	}
	catch (...) {
		MessageDisp(MSG_ER_MEM3_CODE);
		exit(0);
	}
	sizew2 = wcstombs( clipwork, clipdata, sizew2+1);
	clipwork[sizew2] = LIT_NULL_C;
//printf("ペーストした \n%s %d %d\n", clipwork, sizew2, wcslen(clipdata));
	if(data->selection == GDK_SELECTION_CLIPBOARD) {
		if(info == TARGET_STRING) {
//			gtk_selection_data_set(data, GDK_SELECTION_TYPE_STRING,
//	                      8 * sizeof(gchar), (guchar *)clipdata, strlen(clipdata));
			gtk_selection_data_set(data, GDK_SELECTION_TYPE_STRING,
	                      8 * sizeof(gchar), (guchar *)clipwork, sizew2);
//printf("ペースト1\n");
		} else if(info == TARGET_TEXT || info == TARGET_COMPOUND_TEXT) {
			guchar *text;
//printf("ペースト2\n");
			gint format;
			gint length;
			GdkAtom encoding;

			gdk_string_to_compound_text(clipwork, &encoding, &format,
	                            &text, &length);
			gtk_selection_data_set(data, encoding, format, text, length);
			gdk_free_compound_text(text);
		}
	}
//	gtk_main_quit ();
	delete[] clipwork;
}

//==============================================================================
//クリップボードに出力する
//==============================================================================
int ClipWrite(UNICHAR *wkdata) {

	clipdata = wkdata;//ポインタを退避

	//セレクション(クリップボード)の所有を取得する
	gtk_selection_owner_set(window, GDK_SELECTION_CLIPBOARD, GDK_CURRENT_TIME);

	//提供可能なデータのタイプ
	GtkTargetEntry targets[] = {
		{(gchar *)"STRING",        0, TARGET_STRING},
		{(gchar *)"TEXT",          0, TARGET_TEXT}, 
		{(gchar *)"COMPOUND_TEXT", 0, TARGET_COMPOUND_TEXT}};

	g_signal_connect(G_OBJECT(window), "selection-get",
		         G_CALLBACK(signal_selection_out), NULL);

	/* クリップボードで提供可能なデータ形式を登録する */
	gtk_selection_add_targets (window, GDK_SELECTION_CLIPBOARD, targets, G_N_ELEMENTS (targets));

	return 0;
}

//==============================================================================
//セレクション(クリップボード)のデータを取り出し可能になったときのコールバック
//==============================================================================
void signal_selection_received(GtkWidget        *widget,
                                GtkSelectionData *data,
                                guint             time,
                                gpointer          user_data) {

	int irc;
	/* 他のプロセスからデータが送られてきたならばlengthメンバは0以上 */
	if(data->selection == GDK_SELECTION_CLIPBOARD
//#ifdef _GTK_BIT32  //gtk64 long=8byte  gtk32 ling=4byte
//	                      && data->target == (GdkAtom)GPOINTER_TO_INT(user_data)
//#else //_GTK_BIT32
	                      && data->target == (GdkAtom)(glong)(user_data)
//#endif //_GTK_BIT32
	                      && data->type == GDK_SELECTION_TYPE_ATOM
	                      && data->length > 0) {
		gint ia, ib;
		GdkAtom *atoms;
//printf("CLIP %d\n", (int)gdk_selection_owner_get(GDK_SELECTION_PRIMARY));
//printf("CLIP %d %d %d %d\n", (int)data->selection, (int)data->target, (int)data->type, (int)data->length);

		//受け取ったアトムの種類をチェック
		atoms = (GdkAtom *)data->data;
		for(ia = 0; ia < data->length / sizeof(GdkAtom); ia++) {
			gchar *name;
			name = gdk_atom_name(atoms[ia]);
//printf("name %s\n", name);
			if(strncmp(name, clip_key1, strlen(clip_key1)) == 0) {
				for(ib = ia; ib < strlen(name); ib++) {
					if(strncmp(&name[ib], clip_key2, strlen(clip_key2)) == 0) {
//printf("CALC DATA\n");
						text_flag = IN_CALCDATA;
						break;
					}
				}
			} else if(strcmp(name, "UTF8_STRING") == 0) {
#ifndef _UNICODE_TABLE
				if ( text_flag != IN_TEXTDATA ) {
#endif  //_UNICODE_TABLE
					if ( text_flag != IN_CALCDATA ) text_flag = IN_UNIDATA;
#ifndef _UNICODE_TABLE
				}
#endif  //_UNICODE_TABLE
			} else if(strcmp(name, "TEXT") == 0) {
#ifdef _UNICODE_TABLE
				if ( text_flag != IN_UNIDATA ) {
#endif  //_UNICODE_TABLE
					if ( text_flag != IN_CALCDATA ) text_flag = IN_TEXTDATA;
#ifdef _UNICODE_TABLE
				}
#endif  //_UNICODE_TABLE
			} else if(strcmp(name, "STRING") == 0) {
#ifdef _UNICODE_TABLE
				if ( text_flag != IN_UNIDATA ) {
#endif  //_UNICODE_TABLE
					if ( text_flag != IN_CALCDATA ) text_flag = IN_TEXTDATA;
#ifdef _UNICODE_TABLE
				}
#endif  //_UNICODE_TABLE
			} else if(strcmp(name, "COMPOUND_TEXT") == 0) {
#ifdef _UNICODE_TABLE
				if ( text_flag != IN_UNIDATA ) {
#endif  //_UNICODE_TABLE
					if ( text_flag != IN_CALCDATA ) text_flag = IN_TEXTDATA;
#ifdef _UNICODE_TABLE
				}
#endif  //_UNICODE_TABLE
			}
//			if(name) {
//				printf("%s\n", name);
//				g_free(name);
//			} else {
//				printf("不明なアトム:%d\n", (int)atoms[i]);
//			}
		}

//------------------------------------------------------------------------------
//テキストデータなら質問
//------------------------------------------------------------------------------
		if ( text_flag == IN_TEXTDATA || text_flag == IN_UNIDATA ) {
			if ( MessageDisp(MSG_ALL_LINE_CODE) != GTK_RESPONSE_OK ) {
				irc = MessageDisp(MSG_CANCLE_CODE);
				WinClose();
				return;
			} else {
				Allline_flag = 1;
			}
		}
	}

//------------------------------------------------------------------------------
//クリップボードからテキストデータ取り込み
//後でCalcにペーストするときにエラーになったときのために、IN_CALCDATAでもテキストデータを入力しておく
//------------------------------------------------------------------------------
	if ( text_flag == IN_TEXTDATA || text_flag == IN_UNIDATA || text_flag == IN_CALCDATA ) {
		Window_Refresh();
		//EUC非対応
		if (data->selection == GDK_SELECTION_CLIPBOARD && data->length > 0) {
			gint i, count;

//			if (data->target == atom_targets && data->type == GDK_SELECTION_TYPE_ATOM) {
			if (data->type == GDK_SELECTION_TYPE_ATOM) {
//------------------------------------------------------------------------------
//データを要求
//------------------------------------------------------------------------------
				GdkAtom *atoms;

				//TARGETSのとき
				atoms = (GdkAtom *)data->data;
				//データ形式を検索する
				count = data->length / sizeof (GdkAtom);
				for (i = 0; i < count; i++) {
					if (atoms[i] == atom_compound || atoms[i] == atom_text) {
						break;
					}
				}
				//データを要求
				gtk_selection_convert (widget, GDK_SELECTION_CLIPBOARD,
					i < count ? atoms[i] : GDK_TARGET_STRING, GDK_CURRENT_TIME);
				return;
			} else if (data->target == atom_compound || data->target == atom_text) {
//------------------------------------------------------------------------------
//データを取り出す
//------------------------------------------------------------------------------
				gchar **list;

				//コンパウンドテキストのとき
				count = gdk_text_property_to_text_list (data->type, data->format,
						data->data, data->length, &list);
				if (count > 0) {
					for (i = 0; i < count; i++) {
//						printf ("cilpdata %s %d\n", list[i], strlen(list[i]));
//区　点　JIS  SJIS EUC  UTF-8  UTF-16 字
//04 02 2422 82A0 A4A2 E38182 3042   あ
//04 04 2424 82A2 A4A4 E38184 3044   い
					}

					try{
						clip_data = new UNICHAR[strlen(list[0])+1];
					}
#if _MSC_VER > 1200
					catch (std::bad_alloc) {
#else  //MSC
//VC6 以下
					catch (...) {
#endif  //MSC
						irc = MessageDisp( MSG_ER_MEM2_CODE );
						WinClose();
						return;
					}
#if _MSC_VER > 1200
					catch (...) {
						irc = MessageDisp(MSG_ER_MEM3_CODE);
						WinClose();
						return;
					}
#endif  //MSC
					Clip_alloc = 1;
#ifdef _UNICODE_TABLE
					int sizew2 = mbstowcs( clip_data, list[0], strlen(list[0]));
					if ( sizew2 == -1 ) {
						irc = MessageDisp(MSG_ER_CLP9_CODE);
						delete[] clip_data;
						Clip_alloc = 0;
						WinClose();
						return;
					}
					clip_data[strlen(list[0])] = LIT_NULL_C;
#else  //_UNICODE_TABLE
					MTRNCPY ( clip_data, strlen(list[0])+1, (UNICHAR *)list[0], strlen(list[0]) );
					clip_data[strlen(list[0])] = LIT_NULL_C;
#endif  //_UNICODE_TABLE


					gdk_free_text_list (list);
				}

			} else if (data->target == GDK_TARGET_STRING) {
				//文字列のとき
				//ここの取り出し処理は検証していない
				//try{
				//	clip_data = new UNICHAR[strlen(data->data)+1];
				//}
				//MTRNCPY ( clip_data, strlen(data->data), data->data );
			}

//------------------------------------------------------------------------------
//テキストの変換
//------------------------------------------------------------------------------
			Window_Refresh();
//printf("CALC START %ls\n", clip_data);
			if ( CalcGet() != 0 ) {
				WinClose();
				return;//(FALSE);//中止
			}
//printf("CALC END\n");
			WinClose();
			return;//(FALSE);//中止
		}
	}

//------------------------------------------------------------------------------
//データなしなら終わりにする
//------------------------------------------------------------------------------
	if ( text_flag == IN_NOTDATA ) {
		irc = MessageDisp(MSG_ER_CLP8_CODE);
		WinClose();
		return;
	}

//------------------------------------------------------------------------------
//Calcの変換
//------------------------------------------------------------------------------
	if ( text_flag == IN_CALCDATA ) {
		Window_Refresh();
//printf("CALC START\n");
		if ( CalcGet() != 0 ) {
//printf("CALC ERROR END\n");
			WinClose();
			return;//(FALSE);//中止
		}
//printf("CALC END\n");
		WinClose();
		return;//(FALSE);//中止
	}
}

//==============================================================================
//realizeされたときのコールバック　　表示
//==============================================================================
static void signal_realize(GtkWidget *widget, gpointer user_data) {
	time_t aclock;
	int irc;

	time(&aclock);

	Just1 = localtime(&aclock);

//------------------------------------------------------------------------------
//rdbファイルの確認と設定
//------------------------------------------------------------------------------
	if ( Verno >= 1 ) {//3.4以上の場合
		EnvSet();

	} else if ( Verno == 0 ) {//バージョン3.4未満の場合
		if ( RdbCheck() == 0 ) {
			MessageDisp(MSG_ER_RDB_CODE);//rdbファイルが無い
			WinClose();
			return;

		}
	} else {
		MessageDisp( MSG_ER_CAL2_CODE );//Calcが見つからない
		WinClose();
		return;
	}

//------------------------------------------------------------------------------
//クリップボード書き込みの定義
//------------------------------------------------------------------------------
	atom_targets = gdk_atom_intern ("TARGETS", FALSE);
	atom_compound = gdk_atom_intern ("COMPOUND_TEXT", FALSE);
	atom_text = gdk_atom_intern ("TEXT", FALSE);
	/* クリップボードのTARGETSターゲットを要求する */
	gtk_selection_convert (window, GDK_SELECTION_CLIPBOARD, atom_targets,
	                       GDK_CURRENT_TIME);

//------------------------------------------------------------------------------
//クリップボード取得のコールバックの定義
//------------------------------------------------------------------------------
	g_signal_connect(G_OBJECT(window), "selection-received",
	                 G_CALLBACK(signal_selection_received),
	                 GINT_TO_POINTER(atom_targets));
	/* クリップボードのTARGETSターゲットを要求する */
	gtk_selection_convert(window, GDK_SELECTION_CLIPBOARD, atom_targets, GDK_CURRENT_TIME);

}

//==============================================================================
//メイン関数
//==============================================================================
//int main(int argc, char *argv[]) {
SAL_IMPLEMENT_MAIN_WITH_ARGS(argc, argv) {
	GtkWidget *align;
	GtkWidget *table;
	GtkAdjustment *adj;
	GtkWidget *button;
	GtkWidget *vbox;
	GtkWidget *label;
	GdkAtom atom_targets;
	int ia, irc;

	close_button = 0;
	text_flag = IN_NOTDATA;

#ifdef _UNICODE_TABLE
	wsetlocale( LC_CTYPE, LIT_NULL_S );//ロケール設定
#else  //_UNICODE_TABLE
	setlocale( LC_CTYPE, LIT_NULL_S );//ロケール設定
#endif  //_UNICODE_TABLE

//------------------------------------------------------------------------------
//パスの取得
//------------------------------------------------------------------------------
#ifdef _UNICODE_TABLE
	if ( strlen(argv[0]) >= sizeof(AppPath) ) {
		irc = MessageDisp( MSG_ER_PATH_CODE );//バスが長い
		return 0;
	} 
	int sizew2 = mbstowcs( AppPath, argv[0], strlen(argv[0]));
	if ( sizew2 == -1 ) {
		irc = MessageDisp( MSG_ER_PATH_CODE );//バスが不正
		return 0;
	}
	AppPath[strlen(argv[0])] = LIT_NULL_C;
#else  //_UNICODE_TABLE
	MTRCPY( AppPath, sizeof(AppPath), argv[0] );
#endif  //_UNICODE_TABLE
	for(ia=MTRLEN(AppPath); ia>0; ia--) {
		if ( AppPath[ia] == SEP_C ) {
			AppPath[ia] = LIT_NULL_C;
			break;
		}
	}

//------------------------------------------------------------------------------
//初期化ファイルを調べる
//------------------------------------------------------------------------------
#if defined(_WIKI) || defined(_HTML)
	if ( FileSizeGet(AppPath, INIFILE) <= 0 ) {
		irc = MessageDisp( MSG_ER_INIT_CODE );//iniファイルが無い
		Window_Refresh();
	}
#endif  //_WIKI _HTML

	MTRCPY(Inifile, sizeof(Inifile), AppPath);
	MTRCAT(Inifile, sizeof(Inifile), SEP_S);
	MTRCAT(Inifile, sizeof(Inifile), INIFILE);

	irc = ProfileGet(Inifile);
	if ( irc < 0 ) {//コード表が正しくないか読めない
		WinClose();
		return 0;
	}

	if ( MTRCMP( Locale, LIT_NULL_S ) != 0 ) {
#ifdef _UNICODE_TABLE
		wsetlocale( LC_CTYPE, Locale );//ロケール設定
#else  //_UNICODE_TABLE
		setlocale( LC_CTYPE, Locale );//ロケール設定
#endif  //_UNICODE_TABLE
	}

	gtk_set_locale ();
	gtk_init (&argc, &argv);

	Break_flag   = 0;
	Allline_flag = 0;

	SofficePath[0]    = LIT_NULL_C;
	SofficeRdbPath[0] = LIT_NULL_C;

//printf("START %d  %s\n", argc, argv[0]);
//------------------------------------------------------------------------------
//メモリの割り当てなど
//------------------------------------------------------------------------------
	//アトム
	atom_targets = gdk_atom_intern("TARGETS", FALSE);

	//コールバックに渡されるデータ用にメモリを割り当てる
	gdata = (CallbackData*)g_malloc( sizeof(CallbackData) );

	window        = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gdata->window = window;

//------------------------------------------------------------------------------
//LibreOfficeのパスとバージョンチェック
//------------------------------------------------------------------------------
	Verno = Soffice_Path_get();

//------------------------------------------------------------------------------
//ウインドウの設定
//------------------------------------------------------------------------------
	gtk_window_set_default_size(GTK_WINDOW(window), 230, 150);
	gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);

	gtk_window_set_title (GTK_WINDOW (window), APP_NAME);

	gtk_window_set_policy (GTK_WINDOW (window), FALSE, FALSE, TRUE);

	vbox = gtk_vbox_new (FALSE, 3);
	gtk_container_set_border_width (GTK_CONTAINER (vbox), 10);
	gtk_container_add (GTK_CONTAINER (window), vbox);
	gtk_widget_show(vbox);

//------------------------------------------------------------------------------
//ラベルを作る
//------------------------------------------------------------------------------
	label = gtk_label_new (WIN_MSG);
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0);
	gtk_box_pack_start (GTK_BOX (vbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

//------------------------------------------------------------------------------
//プログレスバーを作る
//------------------------------------------------------------------------------
	//中央寄せのアライメントオブジェクトを作成する
	align = gtk_alignment_new (0.5, 0.5, 0, 0);
	gtk_box_pack_start (GTK_BOX (vbox), align, FALSE, FALSE, 5);
	gtk_widget_show(align);

	//プログレスバーの範囲を保持するアジャストメントオブジェクトを作成する
	adj = (GtkAdjustment *) gtk_adjustment_new (0, 1, 100, 0, 0, 0);

	//アジャストメントを用いて GtkProgressBar を作成する
	gdata->pbar = gtk_progress_bar_new_with_adjustment (adj);

	gtk_container_add (GTK_CONTAINER (align), gdata->pbar);
	gtk_widget_show(gdata->pbar);

//------------------------------------------------------------------------------
//キャンセルのボタンを作る
//------------------------------------------------------------------------------
	//rows, columns, homogeneous
	table = gtk_table_new (1, 4, FALSE);//Y分割数 X分割数
	gtk_box_pack_start (GTK_BOX (vbox), table, FALSE, TRUE, 0);

	button = gtk_button_new_with_label (WIN_BTN);
	gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
		                   (GtkSignalFunc) button_push_message,
		                   window);

	gtk_table_attach (GTK_TABLE (table), button, 0, 1, 0, 1,//X左　右 Y上　下
		          GTK_SHRINK, GTK_SHRINK,
                    /*    GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL,*/
//		          GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL,
		              5, 5);

	//ボタンをデフォルトにする
	GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default (button);

	gtk_widget_show(button);
	gtk_widget_show(table);

//------------------------------------------------------------------------------
//関数の定義
//------------------------------------------------------------------------------
	gtk_signal_connect (GTK_OBJECT (window), "delete_event",
		                GTK_SIGNAL_FUNC (cancel_message),
		                gdata);
	gtk_signal_connect (GTK_OBJECT (window), "destroy",
		                GTK_SIGNAL_FUNC (destroy_progress),
		                gdata);
	g_signal_connect(G_OBJECT(window), "realize",
		         G_CALLBACK(signal_realize), NULL);
	gtk_widget_show (window);

//------------------------------------------------------------------------------
//gtk_mainを呼ぶ
//------------------------------------------------------------------------------
	gtk_main ();

	return 0;
}
#endif  //_LINUX_PLAT

