﻿//*****************************************************************************
//* Copyright (c) 2001-2012 Miday Workshop All rights reserved.
//*
//* File: MakeMail.cpp
//*****************************************************************************

#ifdef _WINDOWS_PLAT
#include <windows.h>
#else  //_WINDOWS_PLAT
#include <stdint.h>
#endif  //_WINDOWS_PLAT
#include <new>
#include <stdio.h>
#include <string.h>
#include <wchar.h>

#if defined(_JPN) || defined(_ENG)

#include "CalcDef.h"
#include "SizeDef.h"
#include "MakeMail.h"
#include "MsgLocale.h"

//==============================================================================
//セルの位置と表の大きさを計算する（縦、横方向ともにこの関数で処理）
//==============================================================================
static void Cell_pos_set(POS *pos_p, int *table_wide, int cell_no, int char_wide, int line_wide) {
//    POS *pos_p       入出力　セルの位置データ
//    int *table_wide  出力　　表の幅（高さ）
//    int cell_no      入力　　セルの数
//    int char_wide    入力　　文字の幅のバイト数
//    int line_wide    入力　　線の幅のバイト数
	int ia;
	*table_wide = line_wide;
	for ( ia=0; ia<cell_no; ia++ ) {
		pos_p[ia].posp = *table_wide;
		*table_wide = *table_wide + pos_p[ia].posw * char_wide + line_wide;
//printf("NOX= %d %d %d %d\n", cell_no, *table_wide, pos_p[ia].posw, line_wide);
	}
}

//==============================================================================
//データがあるセルに内の1行に文字列を設定する
//不足分を半角空白で埋めるため、半角を基準に処理を行う
//==============================================================================
static int Line_char_set(UNICHAR *out, int outlen, CELL *hyo_s, int ie, int yoko) {
//    UNICHAR *out     出力　　出力データ
//    int outlen       入力　　出力データ長
//    CELL *hyo_s      入力　　セルデータ
//    int ie           入力　　終了位置
//    int yoko         入力　　セル横幅(ユニコードは文字数、他はバイト数)(unicodeで、AsciiWide=1では * 2した値)
	int charnox, charnox1, charnox2, harfchar, charlen, ma;
	//半角文字数
	harfchar = HalfCheck(hyo_s->text.data, hyo_s->charpos, ie);
	charlen  = ie - hyo_s->charpos + 1;
#ifdef _UNICODE_TABLE
#ifdef _JPN
	charnox =  yoko * 2 - charlen;//yoko * 2は、半角で計算
	if ( AsciiWide == 2 ) {
		harfchar = 0;
	}
	charnox = charnox - (charlen - harfchar);	//全角文字分引く
#else  //_JPN
	charnox =  yoko - charlen;
	charnox = charnox - (charlen - harfchar);	//全角文字分引く
#endif  //_JPN
#else  //_UNICODE_TABLE
	charnox =  yoko - charlen;
#endif  //_UNICODE_TABLE

	charnox1 = 0;//左余白
	charnox2 = 0;//右余白
	if ( charnox > 0 ) {
		if ( hyo_s->h_ichi == CELL_HORI_RIGHT ) {
			charnox1 = charnox;
			charnox2 = 0;
		} else if ( hyo_s->h_ichi == CELL_HORI_LEFT ) {
			charnox1 = 0;
			charnox2 = charnox;
		} else if ( hyo_s->h_ichi == CELL_HORI_CENTER ) {
			charnox1 = charnox / 2;//センターは左側寄りなので切り捨て
			charnox2 = charnox - charnox1;
		} else if ( hyo_s->num_flag == 1 ) {
			charnox1 = charnox;
			charnox2 = 0;
		} else {
			charnox1 = 0;
			charnox2 = charnox;
		}
	}

	if ( charnox1 > 0 ) {
		for(ma=0; ma<charnox1; ma++) {
			out[ma] = LIT_BLANC_C;
		}
		out[charnox1] = LIT_NULL_C;
	}

	MTRNCPY( &out[charnox1], outlen, &hyo_s->text.data[hyo_s->charpos], ie - hyo_s->charpos + 1 );
	out[charnox1 + ie - hyo_s->charpos + 1] = LIT_NULL_C;

	if ( charnox2 > 0 ) {
		for(ma=0; ma<charnox2; ma++) {
			out[charnox1 + ie - hyo_s->charpos + ma + 1] = LIT_BLANC_C;
		}
		out[charnox1 + ie - hyo_s->charpos + charnox2 + 1] = LIT_NULL_C;
	}
	return ie + 1;
}

//==============================================================================
//セルに文字列を設定する
//==============================================================================
static int Cell_char_set( UNICHAR *out, int outlen, CELL *hyo_s, int yoko, int tate) {
//    UNICHAR *out     出力　　出力データ
//    int outlen       入力　　出力データ長
//    CELL *hyo_s      入力　　セルデータ
//    int yoko         入力　　セル横幅(ユニコードは文字数、他はバイト数)
//    int tate         入力　　セル高さ
	int ix, iw;
	int ma, charnoy1, charnoy2;

	charnoy1 = 0;//開始行
	charnoy2 = hyo_s->takasa;//終了行
	if ( tate != hyo_s->takasa ) {
		if ( hyo_s->v_ichi == CELL_VERT_TOP ) {
			charnoy1 = 0;
		} else if ( hyo_s->v_ichi == CELL_VERT_CENTER ) {
			charnoy1 = (tate - hyo_s->takasa + 1)/2;//センターは下側寄りなので+1
		} else {
			charnoy1 = tate - hyo_s->takasa;
		}
		charnoy2 = charnoy1 + hyo_s->takasa;
	}

	out[0] = LIT_NULL_C;
	//空のセルを出力
#if defined(_UNICODE_TABLE) || defined(_ENG)
		iw = yoko*MOJIHABA;
#else  //_UNICODE_TABLE _ENG
		iw =yoko;
#endif  //_UNICODE_TABLE _ENG
	if ( hyo_s->linepos < charnoy1 || hyo_s->linepos > charnoy2 || hyo_s->charpos >= (signed)MTRLEN(hyo_s->text.data) ) {
		for(ma=0; ma<iw; ma++) {
			out[ma] = LIT_BLANC_C;
		}
		out[iw] = LIT_NULL_C;
		hyo_s->linepos = hyo_s->linepos + 1;
		return NOMAL_RETURN;
	}

	for ( ix=hyo_s->charpos+1; ix<(signed)MTRLEN(hyo_s->text.data); ix++ ) {
		if ( hyo_s->text.data[ix] == LIT_HLF ) {
			hyo_s->charpos = Line_char_set(out, outlen, hyo_s, ix-1, yoko)+1;
			hyo_s->linepos = hyo_s->linepos + 1;
			return NOMAL_RETURN;
		}
	}
	if ( hyo_s->text.data[MTRLEN(hyo_s->text.data)-1] != LIT_HLF ) {
		hyo_s->charpos = Line_char_set(out, outlen, hyo_s, MTRLEN(hyo_s->text.data)-1, yoko);
		hyo_s->linepos = hyo_s->linepos + 1;
		return NOMAL_RETURN;
	}
	hyo_s->linepos = hyo_s->linepos + 1;
	return NOMAL_RETURN;
}

//==============================================================================
//クリップボードに出力するデータを編集する
//==============================================================================
int MakeData() {
	int line1, line2, line3, line4, linex, liney;
	int areasize;
#ifdef _WINDOWS_PLAT
	UNICHAR kaigyo[3] = { LIT_HCR, LIT_HLF, 0x00 };
#else  //_WINDOWS_PLAT
	UNICHAR kaigyo[2] = { LIT_HLF, 0x00 };
#endif  //_WINDOWS_PLAT

	int tablesizew, tablesizeh, cellwidth, cellhight;
	int ia, ib, iw, ih, iw2, ih2, irc;
	int ihc, iwc, ihf, ihg;
	int waku_flag;

	UNICHAR *wkarea;
	int wkareasize;

//------------------------------------------------------------------------------
//文字罫線を作成するのに必要な領域を確保する
//AR_FLAG == 1の枠の長さを調べる。もし入らないならその枠を広げる
//------------------------------------------------------------------------------
	tablesizew = 0;
	tablesizeh = 0;
	//横方向
	Cell_pos_set(yoko_p, &tablesizew, Col_no, 2, BYTELEN);

	for ( ih=0; ih<Row_no; ih++ ) {
		for ( iw=0; iw<Col_no; iw++ ) {
			if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 1 ) {
				if ( hyo[ih*Col_no+iw].merge[AR_RIGHT] >= Col_no ) {
					//将来は、エリアの再確保にする
					irc = MessageDisp( MSG_ER_AREA_CODE );
					FreeArea();
					return 3;
				}
				hyo[ih*Col_no+iw].merge[AR_WIDTH] = (hyo[ih*Col_no+iw].merge[AR_RIGHT] - hyo[ih*Col_no+iw].merge[AR_LEFT]) * BYTELEN;
				for ( iw2=hyo[ih*Col_no+iw].merge[AR_LEFT]; iw2<=hyo[ih*Col_no+iw].merge[AR_RIGHT]; iw2++ ) {
					hyo[ih*Col_no+iw].merge[AR_WIDTH] = hyo[ih*Col_no+iw].merge[AR_WIDTH] + yoko_p[iw2].posw;
				}
				if ( hyo[ih*Col_no+iw].merge[AR_WIDTH] < hyo[ih*Col_no+iw].haba ) {
					yoko_p[iw].posw                   = hyo[ih*Col_no+iw].haba + yoko_p[iw].posw - hyo[ih*Col_no+iw].merge[AR_WIDTH];
					hyo[ih*Col_no+iw].merge[AR_WIDTH] = hyo[ih*Col_no+iw].haba;
					Cell_pos_set(yoko_p, &tablesizew, Col_no, 2, BYTELEN);
				}
			}
			hyo[ih*Col_no+iw].charpos = 0;
			hyo[ih*Col_no+iw].linepos = 0;
		}
	}

	//縦方向
	Cell_pos_set(tate_p, &tablesizeh, Row_no, 1, 1);

	for ( ih=0; ih<Row_no; ih++ ) {
		for ( iw=0; iw<Col_no; iw++ ) {
			if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 1 ) {
				if ( hyo[ih*Col_no+iw].merge[AR_BOTTOM] >= Row_no ) {
					//将来は、エリアの再確保にする
					irc = MessageDisp( MSG_ER_AREA_CODE );
					FreeArea();
					return 3;
				}
				hyo[ih*Col_no+iw].merge[AR_HEIGHT] = (hyo[ih*Col_no+iw].merge[AR_BOTTOM] - hyo[ih*Col_no+iw].merge[AR_TOP]);
				for ( ih2=hyo[ih*Col_no+iw].merge[AR_TOP]; ih2<=hyo[ih*Col_no+iw].merge[AR_BOTTOM]; ih2++ ) {
					hyo[ih*Col_no+iw].merge[AR_HEIGHT] = hyo[ih*Col_no+iw].merge[AR_HEIGHT] + tate_p[ih2].posw;
				}
				if ( hyo[ih*Col_no+iw].merge[AR_HEIGHT] < hyo[ih*Col_no+iw].takasa ) {
					tate_p[ih].posw                    = hyo[ih*Col_no+iw].takasa + tate_p[ih].posw - hyo[ih*Col_no+iw].merge[AR_HEIGHT];
					hyo[ih*Col_no+iw].merge[AR_HEIGHT] = hyo[ih*Col_no+iw].takasa;
					Cell_pos_set(tate_p, &tablesizeh, Row_no, 1, 1);
				}
			}
		}
	}

//------------------------------------------------------------------------------
//領域を確保して改行コードを設定する
//------------------------------------------------------------------------------
	// ＡＢＣＤＥ
	// Ｆ　Ｇ　Ｈ
	// ＩＪＫＬＭ
	// Ｎ　Ｏ　Ｐ
	// ＱＲＳＴＵ

	//areasizeは正しくない可能性あり　検証要
	areasize =            tablesizew * (Row_no+1);//ＡＢＣＤＥ ＩＪＫＬＭ ＱＲＳＴＵ
	areasize = areasize + tablesizeh * ((Col_no+1)*BYTELEN+MTRLEN(kaigyo));//ＡＣＥ\n ＦＧＨ\n ＩＫＭ\n ＮＯＰ\n ＱＳＵ\n
	areasize = areasize - (Row_no+1) * (Col_no+1)*BYTELEN - MTRLEN(kaigyo) + 1;//ＡＣＥ ＩＫＭ ＱＳＵ\0
	for ( ih=0; ih<Row_no; ih++ ) {
		for ( iw=0; iw<Col_no; iw++ ) {
			areasize = areasize + hyo[ih*Col_no+iw].charno;
			areasize = areasize + (tate_p[ih].posw - hyo[ih*Col_no+iw].takasa) * hyo[ih*Col_no+iw].haba * 2;    //tate * 2は、半角で計算
			areasize = areasize + (yoko_p[iw].posw - hyo[ih*Col_no+iw].haba) * hyo[ih*Col_no+iw].takasa * 2;    //yoko * 2は、半角で計算
			areasize = areasize + (tate_p[ih].posw - hyo[ih*Col_no+iw].takasa) * (yoko_p[iw].posw - hyo[ih*Col_no+iw].haba) * 2;// * 2は、半角で計算

			//マージの枠線の長さを加える
			if ( hyo[ih*Col_no+iw].merge[AR_FLAG] != 0 && hyo[ih*Col_no+iw].merge[AR_FLAG] != 7 &&
			     hyo[ih*Col_no+iw].merge[AR_FLAG] != 8 && hyo[ih*Col_no+iw].merge[AR_FLAG] != 9 ) {
				areasize = areasize + (yoko_p[iw].posw+1) * 2;//*2は余裕　本当は要らない
			}
			if ( hyo[ih*Col_no+iw].merge[AR_FLAG] != 0 && hyo[ih*Col_no+iw].merge[AR_FLAG] != 3 &&
			     hyo[ih*Col_no+iw].merge[AR_FLAG] != 6 && hyo[ih*Col_no+iw].merge[AR_FLAG] != 9 ) {
				areasize = areasize + (tate_p[ih].posw+1) * 2;//*2は余裕　本当は要らない
			}
		}
	}

	if ( CharLenNewAlloc(&out_data, areasize) == MEMORY_ERROR ) {
		FreeArea();
		return MEMORY_ERROR;
	}
	Out_alloc = 1;
	out_data.data[0] = LIT_NULL_C;

	for ( ih=0; ih<Row_no; ih++ ) {
		for ( ihc=0; ihc<3; ihc++ ) {
			if ( ihc == 1 ) {
				ihf = tate_p[ih].posw;
			} else {
				ihf = 1;
			}
			for ( ihg=0; ihg<ihf; ihg++ ) {
				for ( iw=0; iw<Col_no; iw++ ) {
					for ( iwc=0; iwc<3; iwc++ ) {
						if ( ihc == 0 ) {
							if ( iwc == 0 ) {
								//----------------------------------------------
								//枠の左上　//Ａ(┌) Ｃ(┬) Ｉ(├) Ｋ(┼)
								//----------------------------------------------
								if ( hyo[ih*Col_no+iw].merge[AR_FLAG] != 5 && hyo[ih*Col_no+iw].merge[AR_FLAG] != 6 &&
								     hyo[ih*Col_no+iw].merge[AR_FLAG] != 8 && hyo[ih*Col_no+iw].merge[AR_FLAG] != 9 ) {
									if ( ih == 0 ) {
										if ( iw == 0 ) {//Ａ(┌)
											line1 = hyo[0].line[AR_TOP];
											line2 = hyo[0].line[AR_LEFT];
											if ( line1 == 0 || line2 == 0 ) {
												waku_flag = 0;
											} else {
												waku_flag = line1;                      //上
												if ( waku_flag < line2 ) {
													waku_flag = line2;                  //左
												}
											}
											if ( CharLenNCat(&out_data, &patas[waku_flag][0], BYTELEN) == MEMORY_ERROR ) {//┌(外)
												FreeArea();
												return MEMORY_ERROR;
											}
										} else {//Ｃ(┬)
											line3 = hyo[iw-1].line[AR_TOP];       //  line1┃ line2
											line4 = hyo[iw].line[AR_TOP];         // ━━━╋━━━
											liney = line4 + line3 * LINEPATTERN;  //  line3┃ line4
											line3 = hyo[iw-1].line[AR_RIGHT];
											line4 = hyo[iw].line[AR_LEFT];
											if ( line3 < line4 ) line3 = line4;
											linex = line3;
											if ( CharLenNCat(&out_data, &patac[liney][linex*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
												FreeArea();
												return MEMORY_ERROR;
											}
										}
									} else {
										if ( iw == 0 ) {//Ｉ(├)
											line2 = hyo[(ih-1)*Col_no].line[AR_BOTTOM]; //  line1┃ line2
											line4 = hyo[ih*Col_no].line[AR_TOP];        // ━━━╋━━━
											if ( line2 < line4 ) line2 = line4;         //  line3┃ line4
											liney = line2;
											line2 = hyo[(ih-1)*Col_no].line[AR_LEFT];
											line4 = hyo[ih*Col_no].line[AR_LEFT];
											linex = line4 + line2 * LINEPATTERN;
											if ( CharLenNCat(&out_data, &patac[liney][linex*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
												FreeArea();
												return MEMORY_ERROR;
											}
										} else {//Ｋ(┼)
											line1 = hyo[(ih-1)*Col_no+iw-1].line[AR_BOTTOM]; //  line1┃ line2
											line2 = hyo[(ih-1)*Col_no+iw].line[AR_BOTTOM];   // ━━━╋━━━
											line3 = hyo[ih*Col_no+iw-1].line[AR_TOP];        //  line3┃ line4
											line4 = hyo[ih*Col_no+iw].line[AR_TOP];
											if ( line1 < line3 ) line1 = line3;
											if ( line2 < line4 ) line2 = line4;
											liney = line2 + line1 * LINEPATTERN;
											line1 = hyo[(ih-1)*Col_no+iw-1].line[AR_RIGHT];
											line2 = hyo[(ih-1)*Col_no+iw].line[AR_LEFT];
											line3 = hyo[ih*Col_no+iw-1].line[AR_RIGHT];
											line4 = hyo[ih*Col_no+iw].line[AR_LEFT];
											if ( line1 < line2 ) line1 = line2;
											if ( line3 < line4 ) line3 = line4;
											linex = line3 + line1 * LINEPATTERN;
											if ( CharLenNCat(&out_data, &patac[liney][linex*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
												FreeArea();
												return MEMORY_ERROR;
											}
										}
									}
								}
							} else if ( iwc == 1 ) {
								//----------------------------------------------
								//枠の上　Ｂ(─) Ｄ(─) Ｊ(─) Ｌ(─)
								//----------------------------------------------
								if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 1 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 2 ||
								     hyo[ih*Col_no+iw].merge[AR_FLAG] == 3 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 0 ) {
									waku_flag = hyo[ih*Col_no+iw].line[AR_TOP];
									if ( ih != 0 ) {
										if ( waku_flag < hyo[(ih-1)*Col_no+iw].line[AR_BOTTOM] ) {
											waku_flag = hyo[(ih-1)*Col_no+iw].line[AR_BOTTOM];
										}
									}
									for ( ia=0; ia<yoko_p[iw].posw*(3-HoriWide); ia=ia+BYTELEN ) {
										if ( CharLenNCat(&out_data, &patal[0][waku_flag*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
									}
								} else {
									if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 4 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 7 ) {
										//マージの場合は、ここでもテキストを入れる
										wkareasize = hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_WIDTH] *3;//本当は2倍で可
										if ( wkareasize <= 0 ) wkareasize = 10;
										try{
											wkarea = new UNICHAR[wkareasize];
										}
										catch (...) {
											return MEMORY_ERROR;
										}

										cellwidth = 0;
										for(ib=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_LEFT]; ib<=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_RIGHT]; ib++) {
											cellwidth = cellwidth + yoko_p[ib].posw;
											if ( ib != iw ) cellwidth = cellwidth + BYTELEN;
										}
										cellhight = 0;
										for(ib=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_TOP]; ib<=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_BOTTOM]; ib++) {
											cellhight = cellhight + tate_p[ib].posw;
											if ( ib != ih ) cellhight = cellhight + 1;
										}
										Cell_char_set( wkarea, wkareasize, &hyo[hyo[ih*Col_no+iw].merge[AR_BASE]],
											cellwidth, cellhight);
										if ( CharLenCat(&out_data, wkarea) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
										delete[] wkarea;
									}
								}
							} else if ( iwc == 2 ) {
								//----------------------------------------------
								//枠の右上　Ｅ(┐) Ｍ(┤)  右端のみ
								//----------------------------------------------
								if ( iw == Col_no-1 ) {//一番右のみ
									if ( ih == 0 ) {//Ｅ(┐)
										line1 = hyo[Col_no-1].line[AR_TOP];
										line2 = hyo[Col_no-1].line[AR_RIGHT];
										if ( line1 == 0 || line2 == 0 ) {
											waku_flag = 0;
										} else {
											waku_flag = line1;                      //上
											if ( waku_flag < line2 ) {
												waku_flag = line2;                  //右
											}
										}
										if ( CharLenNCat(&out_data, &patas[waku_flag][BYTELEN], BYTELEN) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
									} else {//Ｍ(┤)
										line1 = hyo[(ih-1)*Col_no+Col_no-1].line[AR_BOTTOM]; //  line1┃ line2
										line3 = hyo[ih*Col_no+Col_no-1].line[AR_TOP];        // ━━━╋━━━
										if ( line1 < line3 ) line1 = line3;                  //  line3┃ line4
										liney = line1 * LINEPATTERN;
										line1 = hyo[(ih-1)*Col_no+Col_no-1].line[AR_RIGHT];
										line3 = hyo[ih*Col_no+Col_no-1].line[AR_RIGHT];
										linex = line3 + line1 * LINEPATTERN;
										if ( CharLenNCat(&out_data, &patac[liney][linex*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
									}
									//改行(kaigyo)を追加
									if ( CharLenCat(&out_data, kaigyo) == MEMORY_ERROR ) {
										FreeArea();
										return MEMORY_ERROR;
									}
								}
							}
						} else if ( ihc == 1 ) {
							if ( iwc == 0 ) {
								//----------------------------------------------
								//枠の左　Ｆ(│) Ｇ(│) Ｎ(│) Ｏ(│)
								//----------------------------------------------
								if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 1 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 4 ||
								     hyo[ih*Col_no+iw].merge[AR_FLAG] == 7 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 0 ) {
									waku_flag = hyo[ih*Col_no+iw].line[AR_LEFT];
									if ( iw != 0 ) {
										if ( waku_flag < hyo[ih*Col_no+iw-1].line[AR_RIGHT] ) {
											waku_flag = hyo[ih*Col_no+iw-1].line[AR_RIGHT];
										}
									}
									if ( CharLenNCat(&out_data, &patal[1][waku_flag*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
										FreeArea();
										return MEMORY_ERROR;
									}
								}
							} else if ( iwc == 1 ) {
								//----------------------------------------------
								//文字列
								//----------------------------------------------
								if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 0 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 1 ||
								     hyo[ih*Col_no+iw].merge[AR_FLAG] == 4 || hyo[ih*Col_no+iw].merge[AR_FLAG] == 7 ) {
									if ( hyo[ih*Col_no+iw].merge[AR_FLAG] == 0 ) {//マージではない場合
										wkareasize = yoko_p[iw].posw * 3;//本当は2倍で可
										if ( wkareasize <= 0 ) wkareasize = 10;
										try{
											wkarea = new UNICHAR[wkareasize];
										}
										catch (...) {
											return MEMORY_ERROR;
										}
										Cell_char_set( wkarea, wkareasize, &hyo[ih*Col_no+iw],
											yoko_p[iw].posw, tate_p[ih].posw);
									} else {//マージの場合
										wkareasize = hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_WIDTH] *3;//本当は2倍で可
										if ( wkareasize <= 0 ) wkareasize = 10;
										try{
											wkarea = new UNICHAR[wkareasize];
										}
										catch (...) {
											return MEMORY_ERROR;
										}

										cellwidth = 0;
										for(ib=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_LEFT]; ib<=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_RIGHT]; ib++) {
											cellwidth = cellwidth + yoko_p[ib].posw;
											if ( ib != iw ) cellwidth = cellwidth + BYTELEN;
										}
										cellhight = 0;
										for(ib=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_TOP]; ib<=hyo[hyo[ih*Col_no+iw].merge[AR_BASE]].merge[AR_BOTTOM]; ib++) {
											cellhight = cellhight + tate_p[ib].posw;
											if ( ib != ih ) cellhight = cellhight + 1;
										}
										Cell_char_set( wkarea, wkareasize, &hyo[hyo[ih*Col_no+iw].merge[AR_BASE]],
											cellwidth, cellhight);
									}
									if ( CharLenCat(&out_data, wkarea) == MEMORY_ERROR ) {
										FreeArea();
										return MEMORY_ERROR;
									}
									delete[] wkarea;
								}
							} else if ( iwc == 2 ) {
								//----------------------------------------------
								//枠の右　Ｈ(│) Ｐ(│)  右端のみ
								//----------------------------------------------
								if ( iw == Col_no-1 ) {//一番右のみ
									waku_flag = hyo[(ih+1)*(Col_no)-1].line[AR_RIGHT];
									if ( CharLenNCat(&out_data, &patal[1][waku_flag*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
										FreeArea();
										return MEMORY_ERROR;
									}
									//改行(kaigyo)を追加
									if ( CharLenCat(&out_data, kaigyo) == MEMORY_ERROR ) {
										FreeArea();
										return MEMORY_ERROR;
									}
								}
							}
						} else if ( ihc == 2 ) {
							if ( ih == Row_no-1 ) {//一番下のみ
								if ( iwc == 0 ) {
									//------------------------------------------
									//表全体の下で枠の左下　Ｑ(└) Ｓ(┴)  下のみ
									//------------------------------------------
									if ( iw == 0 ) {//Ｑ(└)
										line1 = hyo[(Row_no-1)*Col_no].line[AR_BOTTOM];
										line2 = hyo[(Row_no-1)*Col_no].line[AR_LEFT];
										if ( line1 == 0 || line2 == 0 ) {
											waku_flag = 0;
										} else {
											waku_flag = line1;                      //下
											if ( waku_flag < line2 ) {
												waku_flag = line2;                  //左
											}
										}
										if ( CharLenNCat(&out_data, &patas[waku_flag][BYTELEN * 2], BYTELEN) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
									} else {//Ｓ(┴)
										line1 = hyo[(Row_no-1)*Col_no+iw-1].line[AR_BOTTOM]; //  line1┃ line2
										line2 = hyo[(Row_no-1)*Col_no+iw].line[AR_BOTTOM];   // ━━━╋━━━
										liney = line2 + line1 * LINEPATTERN;                 //  line3┃ line4
										line1 = hyo[(Row_no-1)*Col_no+iw-1].line[AR_RIGHT];
										line2 = hyo[(Row_no-1)*Col_no+iw].line[AR_LEFT];
										if ( line1 < line2 ) line1 = line2;
										linex = line1 * LINEPATTERN;
										if ( CharLenNCat(&out_data, &patac[liney][linex*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
									}
								} else if ( iwc == 1 ) {
									//------------------------------------------
									//表全体の下線　Ｒ(─) Ｔ(─)  下のみ
									//------------------------------------------
									waku_flag = hyo[(Row_no-1)*Col_no+iw].line[AR_BOTTOM];
									for ( ia=0; ia<yoko_p[iw].posw*(3-HoriWide); ia=ia+BYTELEN ) {
										if ( CharLenNCat(&out_data, &patal[0][waku_flag*BYTELEN], BYTELEN) == MEMORY_ERROR ) {
											FreeArea();
											return MEMORY_ERROR;
										}
									}
								} else if ( iwc == 2 ) {
									//------------------------------------------
									//表全体の右下　Ｕ(┘)  右下のみ
									//------------------------------------------
									if ( iw == Col_no-1 ) {//一番右のみ
										line1 = hyo[Col_no*Row_no-1].line[AR_BOTTOM];
										line2 = hyo[Col_no*Row_no-1].line[AR_RIGHT];
										if ( line1 == 0 || line2 == 0 ) {
											waku_flag = 0;
										} else {
											waku_flag = line1;                      //下
											if ( waku_flag < line2 ) {
												waku_flag = line2;                  //右
											}
										}
										if ( CharLenNCat(&out_data, &patas[waku_flag][BYTELEN*3], BYTELEN) == MEMORY_ERROR ) { //┘(外)
											FreeArea();
											return MEMORY_ERROR;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}

	return NOMAL_RETURN;
}
#endif  //_ENG _JPN
